/* -*-c-*-
 *
 * $Id: sym.c,v 1.9 1999/10/22 22:36:37 mdw Exp $
 *
 * Symbol table management
 *
 * (c) 1998 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of the mLib utilities library.
 *
 * mLib is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * mLib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with mLib; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------*
 *
 * $Log: sym.c,v $
 * Revision 1.9  1999/10/22 22:36:37  mdw
 * New test structure for symbol tables.
 *
 * Revision 1.8  1999/08/02 14:45:48  mdw
 * Break low-level hashtable code out from sym.
 *
 * Revision 1.7  1999/06/01 09:49:08  mdw
 * Allow things to be looked up by just their caller-supplied hashes.  This
 * actually needs to be thought through better.
 *
 * Revision 1.6  1999/05/26 21:08:31  mdw
 * Rename symbols in line with newer conventions.
 *
 * Revision 1.5  1999/05/13 22:48:37  mdw
 * Twiddle the extension threshold.  Change `-ise' to `-ize' throughout.
 *
 * Revision 1.4  1999/05/06 19:51:35  mdw
 * Reformatted the LGPL notice a little bit.
 *
 * Revision 1.3  1999/05/05 18:50:31  mdw
 * Change licensing conditions to LGPL.
 *
 * Revision 1.2  1998/11/26 19:27:33  mdw
 * Move SYM_NAME into the header file.  Fix bugs.
 *
 * Revision 1.1.1.1  1998/06/17 23:44:42  mdw
 * Initial version of mLib
 *
 */

/*----- Header files ------------------------------------------------------*/

/* --- ANSI headers --- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* --- Local headers --- */

#include "alloc.h"
#include "bits.h"
#include "crc32.h"
#include "exc.h"
#include "hash.h"
#include "sub.h"
#include "sym.h"
#include "track.h"

/*----- Tuning parameters -------------------------------------------------*/

/* --- Initial hash table size --- *
 *
 * This is the initial @mask@ value.  It must be of the form %$2^n - 1$%,
 * so that it can be used to mask of the bottom bits of a hash value.
 */

#define SYM_INITSZ 64			/* Size of a new hash table */

/* --- Maximum load factor --- *
 *
 * This parameter controls how much the table has to be loaded before the
 * table is extended.  The number of elements %$n$%, the number of bins %$b$%
 * and the limit %$l$% satisfy the relation %$n < bl$%; if a new item is
 * added to the table and this relation is found to be false, the table is
 * doubled in size.
 */

#define SYM_LIMIT(n) ((n) * 2)		/* Load factor for growing table */

/*----- Main code ---------------------------------------------------------*/

/* --- @sym_create@ --- *
 *
 * Arguments:	@sym_table *t@ = symbol table to initialize
 *
 * Returns:	---
 *
 * Use:		Initializes the given symbol table.  Raises @EXC_NOMEM@ if
 *		there isn't enough memory.
 */

void sym_create(sym_table *t)
{
  TRACK_CTX("symbol table creation");
  TRACK_PUSH;
  hash_create(&t->t, SYM_INITSZ);
  t->load = SYM_LIMIT(SYM_INITSZ);
  TRACK_POP;
}

/* --- @sym_destroy@ --- *
 *
 * Arguments:	@sym_table *t@ = pointer to symbol table in question
 *
 * Returns:	---
 *
 * Use:		Destroys a symbol table, freeing all the memory it used to
 *		occupy.
 */

void sym_destroy(sym_table *t)
{
  sym_iter i;

  TRACK_CTX("symbol table destruction");
  TRACK_PUSH;

  SYM_MKITER(&i, t);
  for (;;) {
    sym_base *p;
    SYM_NEXT(&i, p);
    if (!p)
      break;
    if (p->len > SYM_BUFSZ)
      sub_free(p->name.p, p->len);
    free(p);
  }
  hash_destroy(&t->t);

  TRACK_POP;
}

/* --- @sym_find@ --- *
 *
 * Arguments:	@sym_table *t@ = pointer to symbol table in question
 *		@const char *n@ = pointer to symbol table to look up
 *		@long l@ = length of the name string or negative to measure
 *		@size_t sz@ = size of desired symbol object, or zero
 *		@unsigned *f@ = pointer to a flag, or null.
 *
 * Returns:	The address of a @sym_base@ structure, or null if not found
 *		and @sz@ is zero.
 *
 * Use:		Looks up a symbol in a given symbol table.  The name is
 *		passed by the address of its first character.  The length
 *		may be given, in which case the name may contain arbitrary
 *		binary data, or it may be given as a negative number, in
 *		which case the length of the name is calculated as
 *		@strlen(n) + 1@.
 *
 *		The return value is the address of a pointer to a @sym_base@
 *		block (which may have other things on the end, as above).  If
 *		the symbol could be found, the return value points to the
 *		symbol block.  If the symbol wasn't there, then if @sz@ is
 *		nonzero, a new symbol is created and its address is returned;
 *		otherwise a null pointer is returned.  The exception
 *		@EXC_NOMEM@ is raised if the block can't be allocated.
 *
 *		The value of @*f@ indicates whether a new symbol entry was
 *		created: a nonzero value indicates that an old value was
 *		found.
 */

void *sym_find(sym_table *t, const char *n, long l, size_t sz, unsigned *f)
{
  uint32 hash;
  size_t len = 0;
  hash_base **bin, **p;
  sym_base *q;

  /* --- Find the correct bin --- */

  len = l < 0 ? strlen(n) + 1 : l;
  CRC32(hash, 0, n, len);
  bin = HASH_BIN(&t->t, hash);

  /* --- Search the bin list --- */

  for (p = bin; *p; p = &(*p)->next) {
    q = (sym_base *)*p;
    if (hash == q->b.hash && len == q->len && !memcmp(n, SYM_NAME(q), len)) {

      /* --- Found a match --- *
       *
       * As a minor, and probably pointless, tweak, move the item to the
       * front of its bin list.
       */

      (*p) = q->b.next;
      q->b.next = *bin;
      *bin = &q->b;

      /* --- Return the block --- */

      if (f) *f = 1;
      return (q);
    }
  }

  /* --- Couldn't find the item there --- */

  if (f) *f = 0;
  if (!sz) return (0);

  /* --- Create a new symbol block and initialize it --- */

  {
    TRACK_CTX("new symbol creation");
    TRACK_PUSH;

    q = xmalloc(sz);
    q->b.next = *bin;
    q->b.hash = hash;
    q->len = len;
    if (n) {
      if (len <= SYM_BUFSZ)
	memcpy(q->name.b, n, len);
      else {
	TRY {
	  q->name.p = sub_alloc(len);
	  memcpy(q->name.p, n, len);
	} CATCH {
	  free(q);
	  TRACK_POP;
	  RETHROW;
	} END_TRY;
      }
    }

    TRACK_POP;
  }

  *bin = &q->b;

  /* --- Consider growing the array --- */

  if (t->load)
    t->load--;
  if (!t->load && hash_extend(&t->t))
    t->load = SYM_LIMIT(t->t.mask / 2 + 1);

  /* --- Finished that, so return the new symbol block --- */

  return (q);
}

/* --- @sym_remove@ --- *
 *
 * Arguments:	@sym_table *i@ = pointer to a symbol table object
 *		@void *p@ = pointer to symbol table entry
 *
 * Returns:	---
 *
 * Use:		Removes the object from the symbol table.  The space occupied
 *		by the object and its name is freed; anything else attached
 *		to the entry should already be gone by this point.
 */

void sym_remove(sym_table *t, void *p)
{
  sym_base *q = p;
  hash_remove(&t->t, &q->b);
  if (q->len > SYM_BUFSZ)
    sub_free(q->name.p, q->len);
  free(q);
  t->load++;
}

/* --- @sym_mkiter@ --- *
 *
 * Arguments:	@sym_iter *i@ = pointer to an iterator object
 *		@sym_table *t@ = pointer to a symbol table object
 *
 * Returns:	---
 *
 * Use:		Creates a new symbol table iterator which may be used to
 *		iterate through a symbol table.
 */

void sym_mkiter(sym_iter *i, sym_table *t) { SYM_MKITER(i, t); }

/* --- @sym_next@ --- *
 *
 * Arguments:	@sym_iter *i@ = pointer to iterator object
 *
 * Returns:	Pointer to the next symbol found, or null when finished.
 *
 * Use:		Returns the next symbol from the table.  Symbols are not
 *		returned in any particular order.
 */

void *sym_next(sym_iter *i)
{
  void *p;
  SYM_NEXT(i, p);
  return (p);
}

/*----- That's all, folks -------------------------------------------------*/
