/* -*-c-*-
 *
 * $Id: sym.h,v 1.9 1999/08/02 16:53:48 mdw Exp $
 *
 * Symbol table management
 *
 * (c) 1998 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of the mLib utilities library.
 *
 * mLib is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * mLib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with mLib; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------*
 *
 * $Log: sym.h,v $
 * Revision 1.9  1999/08/02 16:53:48  mdw
 * Improve type safety for sym_iter objects.
 *
 * Revision 1.8  1999/08/02 14:45:48  mdw
 * Break low-level hashtable code out from sym.
 *
 * Revision 1.7  1999/06/01 09:49:33  mdw
 * Allow things to be looked up by just their caller-supplied hashes.  This
 * actually needs to be thought through better.
 *
 * Revision 1.6  1999/05/26 21:08:31  mdw
 * Rename symbols in line with newer conventions.
 *
 * Revision 1.5  1999/05/13 22:48:37  mdw
 * Change `-ise' to `-ize' throughout.
 *
 * Revision 1.4  1999/05/06 19:51:35  mdw
 * Reformatted the LGPL notice a little bit.
 *
 * Revision 1.3  1999/05/05 18:50:31  mdw
 * Change licensing conditions to LGPL.
 *
 * Revision 1.2  1998/11/26 19:27:34  mdw
 * Move SYM_NAME into the header file.  Fix bugs.
 *
 * Revision 1.1.1.1  1998/06/17 23:44:42  mdw
 * Initial version of mLib
 *
 */

#ifndef SYM_H
#define SYM_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Required headers --------------------------------------------------*/

#include <stddef.h>

#ifndef BITS_H
#  include "bits.h"
#endif

#ifndef HASH_H
#  include "hash.h"
#endif

/*----- Type definitions --------------------------------------------------*/

/* --- Symbol table --- *
 *
 * A @sym_table@ contains the information needed to manage a symbol table.
 * Users shouldn't fiddle with this information directly, but it needs to be
 * here so that objects of the correct type can be created.
 */

typedef struct sym_table {
  hash_table t;
  size_t load;
} sym_table;

/* --- A symbol table entry --- *
 *
 * I don't care what actually gets stored in symbol entries because I don't
 * create them: that's the responsibility of my client.  All I care about
 * here is that whatever gets passed to me is a structure whose first member
 * is a @sym_base@.  The ANSI guarantees about structure layout are
 * sufficient to allow me to manipulate such objects.
 */

#define SYM_BUFSZ 16			/* Size of local string buffer */

typedef struct sym_base {
  hash_base b;
  union {
    char *p;				/* Pointer to name string */
    char b[SYM_BUFSZ];			/* Buffer containing a short name */
  } name;				/* Name of this symbol */
  size_t len;				/* Length of the symbol's name */
} sym_base;

/* --- A macro to pick a symbol's name out from the mess --- */

#define SYM_NAME(sy)							\
  (((sym_base *)(sy))->len > SYM_BUFSZ ?				\
   ((sym_base *)(sy))->name.p :						\
   ((sym_base *)(sy))->name.b)

/* --- An iterator block --- */

typedef struct { hash_iter i; } sym_iter;

/*----- External functions ------------------------------------------------*/

/* --- @sym_create@ --- *
 *
 * Arguments:	@sym_table *t@ = symbol table to initialize
 *
 * Returns:	---
 *
 * Use:		Initializes the given symbol table.  Raises @EXC_NOMEM@ if
 *		there isn't enough memory.
 */

extern void sym_create(sym_table */*t*/);

/* --- @sym_destroy@ --- *
 *
 * Arguments:	@sym_table *t@ = pointer to symbol table in question
 *
 * Returns:	---
 *
 * Use:		Destroys a symbol table, freeing all the memory it used to
 *		occupy.
 */

extern void sym_destroy(sym_table */*t*/);

/* --- @sym_find@ --- *
 *
 * Arguments:	@sym_table *t@ = pointer to symbol table in question
 *		@const char *n@ = pointer to symbol table to look up
 *		@long l@ = length of the name string or negative to measure
 *		@size_t sz@ = size of desired symbol object, or zero
 *		@unsigned *f@ = pointer to a flag, or null.
 *
 * Returns:	The address of a @sym_base@ structure, or null if not found
 *		and @sz@ is zero.
 *
 * Use:		Looks up a symbol in a given symbol table.  The name is
 *		passed by the address of its first character.  The length
 *		may be given, in which case the name may contain arbitrary
 *		binary data, or it may be given as a negative number, in
 *		which case the length of the name is calculated as
 *		@strlen(n) + 1@.
 *
 *		The return value is the address of a pointer to a @sym_base@
 *		block (which may have other things on the end, as above).  If
 *		the symbol could be found, the return value points to the
 *		symbol block.  If the symbol wasn't there, then if @sz@ is
 *		nonzero, a new symbol is created and its address is returned;
 *		otherwise a null pointer is returned.  The exception
 *		@EXC_NOMEM@ is raised if the block can't be allocated.
 *
 *		The value of @*f@ indicates whether a new symbol entry was
 *		created: a nonzero value indicates that an old value was
 *		found.
 */

extern void *sym_find(sym_table */*t*/, const char */*n*/, long /*l*/,
		      size_t /*sz*/, unsigned */*f*/);

/* --- @sym_remove@ --- *
 *
 * Arguments:	@sym_table *i@ = pointer to a symbol table object
 *		@void *b@ = pointer to symbol table entry
 *
 * Returns:	---
 *
 * Use:		Removes the object from the symbol table.  The space occupied
 *		by the object and its name is freed; anything else attached
 *		to the entry should already be gone by this point.
 */

extern void sym_remove(sym_table */*t*/, void */*b*/);

/* --- @sym_mkiter@ --- *
 *
 * Arguments:	@sym_iter *i@ = pointer to an iterator object
 *		@sym_table *t@ = pointer to a symbol table object
 *
 * Returns:	---
 *
 * Use:		Creates a new symbol table iterator which may be used to
 *		iterate through a symbol table.
 */

#define SYM_MKITER(i_, t_) HASH_MKITER(&(i_)->i, &(t_)->t)

extern void sym_mkiter(sym_iter */*i*/, sym_table */*t*/);

/* --- @sym_next@ --- *
 *
 * Arguments:	@sym_iter *i@ = pointer to iterator object
 *
 * Returns:	Pointer to the next symbol found, or null when finished.
 *
 * Use:		Returns the next symbol from the table.  Symbols are not
 *		returned in any particular order.
 */

#define SYM_NEXT(i_, p) do {						\
  hash_base *_q;							\
  HASH_NEXT(&(i_)->i, _q);						\
  (p) = (void *)_q;							\
} while (0)

extern void *sym_next(sym_iter */*i*/);

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
