/* -*-c-*-
 *
 * $Id: testrig.c,v 1.5 1999/05/21 22:14:30 mdw Exp $
 *
 * Generic test driver
 *
 * (c) 1998 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of the mLib utilities library.
 *
 * mLib is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * mLib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with mLib; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: testrig.c,v $
 * Revision 1.5  1999/05/21 22:14:30  mdw
 * Take advantage of the new dynamic string macros.
 *
 * Revision 1.4  1999/05/19 19:02:17  mdw
 * Aesthetic changes: fix spelling of `initialize'; use uppercase token
 * constants; abandon old double-underscore convention for internal
 * functions and variables.
 *
 * Revision 1.3  1999/05/06 19:51:35  mdw
 * Reformatted the LGPL notice a little bit.
 *
 * Revision 1.2  1999/05/05 18:50:31  mdw
 * Change licensing conditions to LGPL.
 *
 * Revision 1.1.1.1  1998/06/17 23:44:42  mdw
 * Initial version of mLib
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dstr.h"
#include "report.h"
#include "quis.h"
#include "testrig.h"

/*----- Static variables --------------------------------------------------*/

static dstr tok;

enum {
  TOK_EOF = 0x100,
  TOK_WORD
};

/*----- Main code ---------------------------------------------------------*/

/* --- @decode@ --- *
 *
 * Arguments:	@int tok@ = token type to decode
 *
 * Returns:	Pointer to a textual representation of the token.
 *
 * Use:		Produces a readable representation of a token.
 */

static const char *decode(int t)
{
  static char buf[4];

  switch (t) {
    case TOK_EOF:
      return ("<eof>");
    case TOK_WORD:
      return (tok.buf);
    default:
      buf[0] = t;
      buf[1] = 0;
      return (buf);
  }
  return ("<buggy-program>");
}

/* --- @gettok@ --- *
 *
 * Arguments:	@FILE *fp@ = file handle to read from
 *
 * Returns:	Type of token read.
 *
 * Use:		Reads a token from the input stream.
 */

static int gettok(FILE *fp)
{
  int ch;

  /* --- Clear the token accumulator --- */

  DRESET(&tok);

  /* --- Prime the lookahead character --- */

again:
  ch = getc(fp);

  /* --- Skip leading whitespace --- */

  while (isspace((unsigned char)ch))
    ch = getc(fp);

  /* --- Trap some special characters --- */

  switch (ch) {

    /* --- Comments --- */

    case '#':
      do ch = getc(fp); while (ch != EOF && ch != '\n');
      goto again;

    /* --- End of file --- */
      
    case EOF:
      return (TOK_EOF);

    /* --- Quote characters --- */

    case '`':
      ch = '\'';
    case '\'':
    case '\"': {
      int quote = ch;

      for (;;) {
	ch = getc(fp);
	if (ch == EOF || ch == quote)
	  break;
	if (ch == '\\') {
	  ch = getc(fp);
	  if (ch == EOF)
	    ch = '\\';
	}
	DPUTC(&tok, ch);
      }
      DPUTZ(&tok);
      return (TOK_WORD);
    }

    /* --- Anything else is either a word or self-delimiting --- */

    default:
      if (isalnum((unsigned char)ch)) {
	for (;;) {
	  DPUTC(&tok, ch);
	  ch = getc(fp);
	  if (ch == EOF ||
	      ch == ';' ||
	      ch == '\"' || ch == '\'' || ch == '`' ||
	      isspace((unsigned char)ch))
	    break;
	  if (ch == '\\') {
	    ch = getc(fp);
	    if (ch == EOF)
	      ch = '\\';
	  }
	}
	ungetc(ch, fp);
	DPUTZ(&tok);
	return (TOK_WORD);
      } else
	return (ch);
  }
}

/* --- @type_hex@ --- */

static void cvt_hex(const char *s, dstr *d)
{
  while (s[0] && s[1]) {
    int x = s[0], y = s[1];
    if ('0' <= x && x <= '9') x -= '0';
    else if ('A' <= x && x <= 'F') x -= 'A' - 10;
    else if ('a' <= x && x <= 'f') x -= 'a' - 10;
    else x = 0;
    if ('0' <= y && y <= '9') y -= '0';
    else if ('A' <= y && y <= 'F') y -= 'A' - 10;
    else if ('a' <= y && y <= 'f') y -= 'a' - 10;
    else y = 0;
    DPUTC(d, (x << 4) + y);
    s += 2;
  }
}

static void dump_hex(dstr *d, FILE *fp)
{
  const char *p, *q;
  for (p = d->buf, q = p + d->len; p < q; p++)
    fprintf(fp, "%02x", *(unsigned char *)p);
}

test_type type_hex = { cvt_hex, dump_hex };

/* --- @type_string@ --- */

static void cvt_string(const char *s, dstr *d)
{
  DPUTS(d, s);
}

static void dump_string(dstr *d, FILE *fp)
{
  DWRITE(d, fp);
}

test_type type_string = { cvt_string, dump_string };

/* --- @type_int@ --- */

static void cvt_int(const char *s, dstr *d)
{
  DENSURE(d, sizeof(long));
  sscanf(s, "%i", (int *)d->buf + d->len);
}

static void dump_int(dstr *d, FILE *fp)
{
  fprintf(fp, "%i", *(int *)(d->buf));
}

test_type type_int = { cvt_int, dump_int };

/* --- @test_run@ --- *
 *
 * Arguments:	@int argc@ = number of command line arguments
 *		@char *argv[]@ = pointer to command line arguments
 *		@const test_chunk chunk[]@ = pointer to chunk definitions
 *		@const char *vec@ = name of default test vector file
 *
 * Returns:	Doesn't.
 *
 * Use:		Runs a set of test vectors to ensure that a component is
 *		working properly.
 */

void test_run(int argc, char *argv[],
	      const test_chunk chunk[],
	      const char *vec)
{
  FILE *fp;
  int i;
  const test_chunk *cch;
  dstr dv[TEST_FIELDMAX];
  int fail = 0, ok = 1;
  int sofar = 0;

  /* --- Silly bits of initialization --- */

  ego(argv[0]);

  for (i = 0; i < TEST_FIELDMAX; i++)
    DCREATE(&dv[i]);

  /* --- Parse command line arguments --- */

  {
    const char *p = 0;

    i = 0;
    for (;;) {
      if (!p || !*p) {
	if (i >= argc - 1)
	  break;
	p = argv[++i];
	if (strcmp(p, "--") == 0) {
	  i++;
	  break;
	}
	if (p[0] != '-' || p[1] == 0)
	  break;
	p++;
      }
      switch (*p++) {
	case 'h':
	  printf("%s test driver\n"
		 "Usage: %s [-f FILENAME]\n", QUIS, QUIS);
	  exit(0);
	case 'f':
	  if (!*p) {
	    if (i >= argc - 1)
	      die(1, "option `-f' expects an argument");
	    p = argv[++i];
	  }
	  vec = p;
	  p = 0;
	  break;
	default:
	  die(1, "option `-%c' unknown", p[-1]);
	  break;
      }
    }
  }

  /* --- Start parsing from the file --- */

  if ((fp = fopen(vec, "r")) == 0)
    die(1, "couldn't open test vector file `%s': %s", vec, strerror(errno));

  for (;;) {
    int t = gettok(fp);

    /* --- This is a reasonable place to stop --- */

    if (t == TOK_EOF)
      break;

    /* --- Pick out the chunk name --- */

    if (t != TOK_WORD)
      die(1, "expected <word>; found `%s'", decode(t));

    /* --- Find the right chunk block --- */

    for (cch = chunk; ; cch++) {
      if (!cch->name)
	goto skip_chunk;
      if (strcmp(tok.buf, cch->name) == 0)
	break;
    }

    /* --- Past the open brace to the first chunk --- */

    if ((t = gettok(fp)) != '{')
      die(1, "expected '{'; found `%s'", decode(t));

    /* --- Start on the test data now --- */

    printf("%s: ", cch->name);
    fflush(stdout);
    sofar = 0;
    ok = 1;

    for (;;) {
      t = gettok(fp);

      /* --- Accept a close brace --- */

      if (t == '}')
	break;

      /* --- Otherwise I expect a list of words --- */

      for (i = 0; cch->f[i]; i++) {
	DRESET(&dv[i]);
	if (t != TOK_WORD)
	  die(1, "expected <word>; found `%s'", decode(t));
	cch->f[i]->cvt(tok.buf, &dv[i]);
	t = gettok(fp);
      }

      /* --- And a terminating semicolon --- */

      if (t != ';')
	die(1, "expected `;'; found `%s'", decode(t));

      /* --- Run the test code --- */

      if (!cch->test(dv)) {
	printf("%s: ", cch->name);
	for (i = 0; i < sofar; i++) putchar('.');
	fail = 1; ok = 0;
      }
      sofar++;
      putchar('.');
      fflush(stdout);
    }

    puts(ok ? " ok" : " failed");
    fflush(stdout);
    continue;

  skip_chunk:
    if ((t = gettok(fp)) != '{')
      die(1, "expected '{'; found `%s'", decode(t));
    for (;;) {
      t = gettok(fp);
      if (t == '}')
	break;
      while (t == TOK_WORD)
	t = gettok(fp);
      if (t != ';')
	die(1, "expected `;'; found `%s'", decode(t));
    }      
  }

  exit(fail);
}

/*----- That's all, folks -------------------------------------------------*/
